/*
 * Copyright (C) 2025 Muhammad Tayyab Akram
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <API/SBBase.h>
#include <Core/ThreadFence.h>

#include "Once.h"

SB_INTERNAL SBBoolean OnceTryExecute(OnceRef once, void(*func)(void *info), void *info)
{
    SBBoolean result = SBTrue;

    if (!once->executed) {
        result = AtomicFlagTestAndSet(&once->flag) == SBFalse;

        if (result) {
            func(info);
            ThreadFence();
            once->executed = SBTrue;
        }
    }

    return result;
}
